/*
 * This software is copyright 1992 by Robert Morris.
 * You may freely redistribute this software as shareware
 * if you do so in the same form as you got it. If you find
 * this software useful, please send $12 to:
 *   Robert Morris
 *   P.O. Box 1044
 *   Harvard Square Station
 *   Cambridge, MA 02238
 *   ecognome@aol.com
 * If you incorporate any of this software in any kind of
 * commercial product, please send $2 per copy distributed
 * to the above address.
 */

#include "bits.h"

/* table to reverse bits in a byte */
static unsigned char reverse[];

#if 0
/*
 * given a byte, return the position of the first 1. low bit is #0.
 * -1 means none.
 */
static signed char first1[];
#endif

OSErr
NewBits(struct bits *b, unsigned char *data, long nbits, char bigendian, int writeable)
{
	b->bits = data;
	b->nbits = nbits;
	b->nextbit = 0;
	b->bigendian = bigendian;
	b->cached = 0;
	if(writeable && bigendian == 0)
		return(-1);
	return(0);
}

OSErr
FillBitCache(struct bits *b)
{
	long nbytes, bitno;
	long byteno, cache;
	
	if(TellBits(b) < 0 || TellBits(b) >= b->nbits)
		return(-1);

	if((b->nextbit & 7) != 0){
		/* user must have called SeekBits(), so cache is no good */
		byteno = b->nextbit >> 3;
		bitno = b->nextbit & 7;
		
		b->cached = 8 - bitno;
		if(b->cached + b->nextbit > b->nbits)
			b->cached = b->nbits - b->nextbit;
		b->nextbit += b->cached;
		
		cache = b->bits[byteno];
		if(b->bigendian)
			cache = reverse[cache];
		cache >>= bitno;
		b->cache = cache << 1;
	} else if(b->cached < 24){
		b->cache &= (1 << (b->cached + 1)) - 1;
	}
	
	/* now b->nextbit is aligned to a byte */
	
	nbytes = ((24 - b->cached) + 7) / 8;
	byteno = b->nextbit >> 3;
	while(--nbytes >= 0){
		cache = b->bits[byteno];
		if(b->bigendian)
			cache = reverse[cache];
		b->cache |= (cache << (b->cached + 1));
		b->cached += 8;
		b->nextbit += 8;
		byteno++;
	}
	
	return(0);
}

long
_ReadBit(struct bits *b)
{
	if(FillBitCache(b) < 0)
		return(-1);	
	return(ReadBit(b));
}

#if 0
/*
 * Search, starting with the current bit, for a bit with the same value as wanted.
 * Tries to do byte-at-a-time operations; optimized for longish runs of the same
 * value. A return of -1 means it reached end of file without finding the bit.
 * Otherwise it leaves the position one after the desired bit.
 */
OSErr
ScanBits(struct bits *b, long wanted)
{
	unsigned char *p, *endp;
	int nleft, first;
	unsigned char cache;
	
	/* move to a byte boundary */
	nleft = 8 - (TellBits(b) & 7);
	if(nleft > b->nbits - TellBits(b))
		nleft = b->nbits - TellBits(b);
	if(nleft != 8){
		if(nleft == b->cached){
			cache = b->cache >> 1;
			if(wanted == 0)
				cache = (~cache & ((1 << nleft) - 1));
			first = first1[cache];
			if(first != -1){
				SeekBits(b, TellBits(b) + first + 1);
				return(b->nextbit <= b->nbits ? 0 : -1);
			} else {
				SeekBits(b, TellBits(b) + nleft);
			}
		} else {
			while(--nleft >= 0)
				if(ReadBit(b) == wanted)
					return(b->nextbit <= b->nbits ? 0 : -1);
		}
	}
		
	/* now we can examine a byte at a time */
	p = b->bits + (TellBits(b) >> 3);
	endp = b->bits + ((b->nbits + 7) >> 3);
	if(wanted){
		while(p < endp && *p == 0)
			p++;
	} else {
		while(p < endp && *p == 0xff)
			p++;
	}
	
	SeekBits(b, (p - b->bits) * 8);
	
	/* now look inside this final byte */
	nleft = b->nbits - TellBits(b);
	while(--nleft >= 0)
		if(ReadBit(b) == wanted)
			return(0);
			
	return(-1);
}
#endif

OSErr
WriteBits(struct bits *b, long value, long n)
{
	long nhere, here, bitno;
	long byteno;
	
	if(n < 0 || n > 32 || b->nextbit + n > b->nbits)
		return(-1);
#if 0
	/* checked in NewBits */
	if(b->bigendian == 0)
		return(-1);
#endif
		
	byteno = b->nextbit >> 3;
	bitno = b->nextbit & 7;
	b->nextbit += n;
	
	while(n > 0){
		/* how many bits can we write to this byte? */
		nhere = 8 - bitno;
		if(nhere > n)
			nhere = n;
			
		here = (value >> (n - nhere)) & ((1 << nhere) - 1);
		
		if(nhere == 8){
			b->bits[byteno] = here;
		} else {
			here <<= (8 - bitno) - nhere;
			b->bits[byteno] = (b->bits[byteno] & (0xff << (8 - bitno)))
							| here;
		}
			
		n -= nhere;
		bitno += nhere;
		if(bitno >= 8){
			bitno = 0;
			byteno += 1;
		}
	}
	
	return(0);
}

/*
 * Write n bits, all of the same value.
 */
OSErr
_WriteBitSpan(struct bits *b, long bit, long n)
{
	long bitno;
	unsigned char *p, *endp;
	long start = b->nextbit;
	
	if(n < 0 || b->nextbit + n > b->nbits)
		return(-1);
#if 0
	/* checked in NewBits */
	if(b->bigendian == 0)
		return(-1);
#endif
	b->nextbit += n;
		
	if(bit)
		bit = 0xff;
		
	p = b->bits + (start >> 3);
		
	/* align to byte boundary */
	bitno = start & 7;
	if(bitno){
		*p = (bit >> bitno) | (*p & (0xff00 >> bitno));
		n -= 8 - bitno;
		if(n <= 0)
			return(0);
		p++;
	}
	
	/* now write whole bytes */
	endp = p + (n >> 3);
	if(bit){
		while(p < endp)
			*p++ = 0xff;
	} else {
		while(p < endp)
			*p++ = 0;
	}
	
	if(n & 7)
		*p = bit;
	
	return(0);
}

static unsigned char reverse[] = {
0x00, 0x80, 0x40, 0xc0, 0x20, 0xa0, 0x60, 0xe0, 
0x10, 0x90, 0x50, 0xd0, 0x30, 0xb0, 0x70, 0xf0, 
0x08, 0x88, 0x48, 0xc8, 0x28, 0xa8, 0x68, 0xe8, 
0x18, 0x98, 0x58, 0xd8, 0x38, 0xb8, 0x78, 0xf8, 
0x04, 0x84, 0x44, 0xc4, 0x24, 0xa4, 0x64, 0xe4, 
0x14, 0x94, 0x54, 0xd4, 0x34, 0xb4, 0x74, 0xf4, 
0x0c, 0x8c, 0x4c, 0xcc, 0x2c, 0xac, 0x6c, 0xec, 
0x1c, 0x9c, 0x5c, 0xdc, 0x3c, 0xbc, 0x7c, 0xfc, 
0x02, 0x82, 0x42, 0xc2, 0x22, 0xa2, 0x62, 0xe2, 
0x12, 0x92, 0x52, 0xd2, 0x32, 0xb2, 0x72, 0xf2, 
0x0a, 0x8a, 0x4a, 0xca, 0x2a, 0xaa, 0x6a, 0xea, 
0x1a, 0x9a, 0x5a, 0xda, 0x3a, 0xba, 0x7a, 0xfa, 
0x06, 0x86, 0x46, 0xc6, 0x26, 0xa6, 0x66, 0xe6, 
0x16, 0x96, 0x56, 0xd6, 0x36, 0xb6, 0x76, 0xf6, 
0x0e, 0x8e, 0x4e, 0xce, 0x2e, 0xae, 0x6e, 0xee, 
0x1e, 0x9e, 0x5e, 0xde, 0x3e, 0xbe, 0x7e, 0xfe, 
0x01, 0x81, 0x41, 0xc1, 0x21, 0xa1, 0x61, 0xe1, 
0x11, 0x91, 0x51, 0xd1, 0x31, 0xb1, 0x71, 0xf1, 
0x09, 0x89, 0x49, 0xc9, 0x29, 0xa9, 0x69, 0xe9, 
0x19, 0x99, 0x59, 0xd9, 0x39, 0xb9, 0x79, 0xf9, 
0x05, 0x85, 0x45, 0xc5, 0x25, 0xa5, 0x65, 0xe5, 
0x15, 0x95, 0x55, 0xd5, 0x35, 0xb5, 0x75, 0xf5, 
0x0d, 0x8d, 0x4d, 0xcd, 0x2d, 0xad, 0x6d, 0xed, 
0x1d, 0x9d, 0x5d, 0xdd, 0x3d, 0xbd, 0x7d, 0xfd, 
0x03, 0x83, 0x43, 0xc3, 0x23, 0xa3, 0x63, 0xe3, 
0x13, 0x93, 0x53, 0xd3, 0x33, 0xb3, 0x73, 0xf3, 
0x0b, 0x8b, 0x4b, 0xcb, 0x2b, 0xab, 0x6b, 0xeb, 
0x1b, 0x9b, 0x5b, 0xdb, 0x3b, 0xbb, 0x7b, 0xfb, 
0x07, 0x87, 0x47, 0xc7, 0x27, 0xa7, 0x67, 0xe7, 
0x17, 0x97, 0x57, 0xd7, 0x37, 0xb7, 0x77, 0xf7, 
0x0f, 0x8f, 0x4f, 0xcf, 0x2f, 0xaf, 0x6f, 0xef, 
0x1f, 0x9f, 0x5f, 0xdf, 0x3f, 0xbf, 0x7f, 0xff, 
};

#if 0
static signed char first1[] = {
-1, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
5, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
6, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
5, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
7, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
5, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
6, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
5, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
};
#endif